import os
from pathlib import Path
from typing import Optional, Sequence, Union

from dotenv import find_dotenv as _find_dotenv


def find_dotenv(
    filename: str = '.env',
    raise_error_if_not_found: bool = False,
    usecwd: bool = False,
    extra_paths: Optional[Sequence[Union[Path, str]]] = None,
) -> str:
    """Find the path to a relevant dotenv file, or return an empty string.

    Extends :py:func:`dotenv.find_dotenv` by also searching in the execution
    path and their parent directories (e.g. virtualenv directories).
    """
    try:
        dotenv_path = _find_dotenv(filename, raise_error_if_not_found, usecwd)
    except IOError:
        # Reraise this error later if our own search fails
        if raise_error_if_not_found:
            pass
    # If find_dotenv doesn't find a dotenv file, check the execution paths
    # (and extra paths if provided) and their parent directories
    if not dotenv_path:
        paths = [Path(p) for p in extra_paths] if extra_paths else []
        paths += [Path(p) for p in os.get_exec_path()]
        for path in paths:
            while path.as_posix() != '/' and not dotenv_path:
                if (path / filename).is_file():
                    dotenv_path = (path / filename).as_posix()
                path = path.parent
            if dotenv_path:
                break
    if not dotenv_path and raise_error_if_not_found:
        raise IOError('File not found')
    return dotenv_path



from importlib import import_module
from importlib.machinery import SourceFileLoader
from importlib.util import spec_from_loader, module_from_spec
from pathlib import Path
import sys
from types import ModuleType
from typing import List, Union

import django
from django.core.management.utils import get_random_secret_key

# Relative path to settings.py template from django package root dir
RELPATH_TO_SETTINGS_TPL = 'conf/project_template/project_name/settings.py-tpl'

# Settings which are in the settings.py template, but we ignore
TEMPLATE_SETTINGS_IGNORE = (
    'BASE_DIR',   # Will find an incorrect dir, so omit
    'DATABASES',  # Depends on BASE_DIR, see above
    'DEBUG',      # Defaults to True - we force it to False
    'SECRET_KEY'  # Needs to be generated
)


def find_paths_to_project_root() -> List[Path]:
    """Try to find the project root without knowing the project name."""
    for arg in sys.argv:
        if arg.endswith('.asgi:application') or arg.endswith('.wsgi'):
            try:
                return [get_base_dir(arg.split('.')[0])]
            except ModuleNotFoundError:
                pass
        if arg.endswith('manage.py'):
            return [Path(arg).parent.absolute()]
    return []


def get_base_dir(project_name: str) -> Path:
    return Path(import_module(project_name).__file__).parent.parent.absolute()


def get_module_settings(project_name: str) -> dict:
    try:
        mod = import_module(f'{project_name}.settings')
        setting_names = [s for s in dir(mod) if s.isupper()]
        return {s: getattr(mod, s) for s in setting_names}
    except ModuleNotFoundError:
        return {}


def get_path_to_settings_template() -> Path:
    return Path(django.__file__).parent / RELPATH_TO_SETTINGS_TPL


def get_template_settings(project_name: str) -> dict:
    template_settings = import_module_from_file(
        get_path_to_settings_template()
    )
    setting_names = [s for s in dir(template_settings) if s.isupper()
                     and s not in TEMPLATE_SETTINGS_IGNORE]
    settings = {}
    for s in setting_names:
        val = getattr(template_settings, s)
        if isinstance(val, str) and '{{ project_name }}' in val:
            val = val.replace('{{ project_name }}', project_name)
        settings[s] = val
    # Set some defaults of our own
    settings['BASE_DIR'] = get_base_dir(project_name)
    settings['DEBUG'] = False  # Ensure DEBUG defaults to False
    settings['SECRET_KEY'] = get_random_secret_key()
    return settings


def import_module_from_file(path: Union[Path, str]) -> ModuleType:
    """Import a Python module based on file path rather than a dotted
    name.
    """
    if isinstance(path, str):
        path = Path(path)
    name = path.name.split('.')[0]
    spec = spec_from_loader(name, SourceFileLoader(name, path.as_posix()))
    mod = module_from_spec(spec)
    spec.loader.exec_module(mod)  # type: ignore  # noqa
    return mod



from typing import Dict, List

# NOTE(elyas): pydantic (or similar) may offer a more robust method of
# defining setting types. For example, JSON does not support storing Python
# tuples, but settings like ADMINS (i.e. List[Tuple[str, str]]) may require
# list elements to be cast to tuple.
_setting_types: Dict[str, List[type]] = {
    'ABSOLUTE_URL_OVERRIDES': [dict],
    'ADMINS': [list],
    'ALLOWED_HOSTS': [list],
    'APPEND_SLASH': [bool],
    'AUTHENTICATION_BACKENDS': [list],
    'AUTH_PASSWORD_VALIDATORS': [list],
    'AUTH_USER_MODEL': [str],
    'CACHES': [dict],
    'CACHE_MIDDLEWARE_ALIAS': [str],
    'CACHE_MIDDLEWARE_KEY_PREFIX': [str],
    'CACHE_MIDDLEWARE_SECONDS': [int],
    'CSRF_COOKIE_AGE': [type(None), int],
    'CSRF_COOKIE_DOMAIN': [type(None), str],
    'CSRF_COOKIE_HTTPONLY': [bool],
    'CSRF_COOKIE_NAME': [str],
    'CSRF_COOKIE_PATH': [str],
    'CSRF_COOKIE_SAMESITE': [str],  # 'None' is a valid str value, so don't coerce to NoneType  # noqa
    'CSRF_COOKIE_SECURE': [bool],
    'CSRF_FAILURE_VIEW': [str],
    'CSRF_HEADER_NAME': [str],
    'CSRF_TRUSTED_ORIGINS': [list],
    'CSRF_USE_SESSIONS': [bool],
    'DATABASES': [dict],
    'DATABASE_ROUTERS': [list],
    'DATA_UPLOAD_MAX_MEMORY_SIZE': [type(None), int],
    'DATA_UPLOAD_MAX_NUMBER_FIELDS': [type(None), int],
    'DATETIME_FORMAT': [str],
    'DATETIME_INPUT_FORMATS': [list],
    'DATE_FORMAT': [str],
    'DATE_INPUT_FORMATS': [list],
    'DEBUG': [bool],
    'DEBUG_PROPAGATE_EXCEPTIONS': [bool],
    'DECIMAL_SEPARATOR': [str],
    'DEFAULT_CHARSET': [str],
    'DEFAULT_CONTENT_TYPE': [str],  # deprecated v2.0, removed v3.0
    'DEFAULT_EXCEPTION_REPORTER': [str],
    'DEFAULT_EXCEPTION_REPORTER_FILTER': [str],
    'DEFAULT_FILE_STORAGE': [str],
    'DEFAULT_FROM_EMAIL': [str],
    'DEFAULT_HASHING_ALGORITHM': [str],
    'DEFAULT_INDEX_TABLESPACE': [str],
    'DEFAULT_TABLESPACE': [str],
    'DISALLOWED_USER_AGENTS': [list],
    'EMAIL_BACKEND': [str],
    'EMAIL_HOST': [str],
    'EMAIL_HOST_PASSWORD': [str],
    'EMAIL_HOST_USER': [str],
    'EMAIL_PORT': [int],
    'EMAIL_SSL_CERTFILE': [type(None), str],
    'EMAIL_SSL_KEYFILE': [type(None), str],
    'EMAIL_SUBJECT_PREFIX': [str],
    'EMAIL_TIMEOUT': [type(None), int],
    'EMAIL_USE_LOCALTIME': [bool],
    'EMAIL_USE_SSL': [bool],
    'EMAIL_USE_TLS': [bool],
    'FILE_CHARSET': [str],  # deprecated v2.2, removed v3.1
    'FILE_UPLOAD_DIRECTORY_PERMISSIONS': [type(None), str],
    'FILE_UPLOAD_HANDLERS': [list],
    'FILE_UPLOAD_MAX_MEMORY_SIZE': [int],
    'FILE_UPLOAD_PERMISSIONS': [type(None), int],
    'FILE_UPLOAD_TEMP_DIR': [type(None), str],
    'FIRST_DAY_OF_WEEK': [int],
    'FIXTURE_DIRS': [list],
    'FORCE_SCRIPT_NAME': [type(None), str],
    'FORMAT_MODULE_PATH': [type(None), str, list],
    'FORM_RENDERER': [str],
    'IGNORABLE_404_URLS': [list],
    'INSTALLED_APPS': [list],
    'INTERNAL_IPS': [list],
    'LANGUAGES': [list],
    'LANGUAGES_BIDI': [list],
    'LANGUAGE_CODE': [str],
    'LANGUAGE_COOKIE_AGE': [type(None), int],
    'LANGUAGE_COOKIE_DOMAIN': [type(None), str],
    'LANGUAGE_COOKIE_HTTPONLY': [bool],
    'LANGUAGE_COOKIE_NAME': [str],
    'LANGUAGE_COOKIE_PATH': [str],
    'LANGUAGE_COOKIE_SAMESITE': [type(None), str],
    'LANGUAGE_COOKIE_SECURE': [bool],
    'LOCALE_PATHS': [list],
    'LOGGING': [dict],
    'LOGGING_CONFIG': [type(None), str],
    'LOGIN_REDIRECT_URL': [str],
    'LOGIN_URL': [str],
    'LOGOUT_REDIRECT_URL': [type(None), str],
    'MANAGERS': [list],
    'MEDIA_ROOT': [str],
    'MEDIA_URL': [str],
    'MESSAGE_STORAGE': [str],
    'MIDDLEWARE': [type(None), list],
    'MIDDLEWARE_CLASSES': [list],  # deprecated v1.10, removed v2.0
    'MIGRATION_MODULES': [dict],
    'MONTH_DAY_FORMAT': [str],
    'NUMBER_GROUPING': [int],
    'PASSWORD_HASHERS': [list],
    'PASSWORD_RESET_TIMEOUT': [int],
    'PASSWORD_RESET_TIMEOUT_DAYS': [int],
    'PREPEND_WWW': [bool],
    'SECRET_KEY': [str],
    'SECURE_BROWSER_XSS_FILTER': [bool],
    'SECURE_CONTENT_TYPE_NOSNIFF': [bool],
    'SECURE_HSTS_INCLUDE_SUBDOMAINS': [bool],
    'SECURE_HSTS_PRELOAD': [bool],
    'SECURE_HSTS_SECONDS': [int],
    'SECURE_PROXY_SSL_HEADER': [type(None), tuple],
    'SECURE_REDIRECT_EXEMPT': [list],
    'SECURE_REFERRER_POLICY': [type(None), str],  # v3.0 default None, v3.1 default str  # noqa
    'SECURE_SSL_HOST': [type(None), str],
    'SECURE_SSL_REDIRECT': [bool],
    'SERVER_EMAIL': [str],
    'SESSION_CACHE_ALIAS': [str],
    'SESSION_COOKIE_AGE': [int],
    'SESSION_COOKIE_DOMAIN': [type(None), str],
    'SESSION_COOKIE_HTTPONLY': [bool],
    'SESSION_COOKIE_NAME': [str],
    'SESSION_COOKIE_PATH': [str],
    'SESSION_COOKIE_SAMESITE': [str, bool],  # 'None' is a valid str value, so don't coerce to NoneType  # noqa
    'SESSION_COOKIE_SECURE': [bool],
    'SESSION_ENGINE': [str],
    'SESSION_EXPIRE_AT_BROWSER_CLOSE': [bool],
    'SESSION_FILE_PATH': [type(None), str],
    'SESSION_SAVE_EVERY_REQUEST': [bool],
    'SESSION_SERIALIZER': [str],
    'SHORT_DATETIME_FORMAT': [str],
    'SHORT_DATE_FORMAT': [str],
    'SIGNING_BACKEND': [str],
    'SILENCED_SYSTEM_CHECKS': [list],
    'STATICFILES_DIRS': [list],
    'STATICFILES_FINDERS': [list],
    'STATICFILES_STORAGE': [str],
    'STATIC_ROOT': [type(None), str],
    'STATIC_URL': [type(None), str],
    'TEMPLATES': [list],
    # 'TEMPLATE_DIRS': [list],  # deprecated v1.8, removed v1.10
    'TEST_NON_SERIALIZED_APPS': [list],
    'TEST_RUNNER': [str],
    'THOUSAND_SEPARATOR': [str],
    'TIME_FORMAT': [str],
    'TIME_INPUT_FORMATS': [list],
    'TIME_ZONE': [type(None), str],
    'USE_ETAGS': [bool],  # deprecated v1.11, removed v2.1
    'USE_I18N': [bool],
    'USE_L10N': [bool],
    'USE_THOUSAND_SEPARATOR': [bool],
    'USE_TZ': [bool],
    'USE_X_FORWARDED_HOST': [bool],
    'USE_X_FORWARDED_PORT': [bool],
    'WSGI_APPLICATION': [type(None), str],
    'X_FRAME_OPTIONS': [str],
    'YEAR_MONTH_FORMAT': [str],
}


def get_setting_types():
    return _setting_types.copy()



from pathlib import Path

from setuptools import find_packages, setup

with open(Path(__file__).parent / 'README.md') as f:
    long_description = f.read()

setup(
    name='django-envconfig',
    packages=find_packages(exclude=['tests']),
    version='0.1.2',
    license='MIT',
    description='Configure Django using environment variables.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Elyas Khan',
    author_email='mail@ely.as',
    url='https://github.com/ely-as/django-envconfig',
    project_urls={
        'Source': 'https://github.com/ely-as/django-envconfig',
        'Tracker': 'https://github.com/ely-as/django-envconfig/issues',
    },
    keywords=[],
    install_requires=[
        'Django',
        'python-dotenv',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Typing :: Typed',
        'Framework :: Django',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 2.0',
        'Framework :: Django :: 2.1',
        'Framework :: Django :: 2.2',
        'Framework :: Django :: 3.0',
        'Framework :: Django :: 3.1',
    ],
)



from os import environ, getenv
import sys

from django.core.exceptions import ImproperlyConfigured
from dotenv import load_dotenv

from envconfig.dotenv import find_dotenv
from envconfig.parser import EnvParser
from envconfig.setting_types import get_setting_types
from envconfig import utils

extra_paths = utils.find_paths_to_project_root()
load_dotenv(dotenv_path=find_dotenv(extra_paths=extra_paths))

project_name = getenv('DJANGO_PROJECT')
if not project_name:
    raise ImproperlyConfigured("Required environment variable "
                               "'DJANGO_PROJECT' not found.")

# Generate default settings and values from startproject template settings.py
settings = utils.get_template_settings(project_name)
# Load any settings from the project_name.settings module (if it exists)
mod_settings = utils.get_module_settings(project_name)
# Overlay the default template settings with any settings module settings
settings.update(mod_settings)

# Get a lookup of built-in Django settings and their valid types
setting_types = get_setting_types()
# Add any custom settings and their default types
setting_types.update({
    s: [type(mod_settings[s])] for s in mod_settings if s not in setting_types
})

if project_name not in settings['INSTALLED_APPS']:
    settings['INSTALLED_APPS'].append(project_name)

envsetting_names = [s for s in setting_types if s in environ]

for name in envsetting_names:
    val = getenv(name)
    if val:
        try:
            settings[name] = EnvParser.parse(val, *setting_types[name])
        except ValueError as e:
            types_str = "', '".join(t.__name__ for t in setting_types[name])
            raise ImproperlyConfigured(
                f"Environment variable '{name}' incorrectly set. "
                f"Error: {str(e)}. Valid types include: '{types_str}'."
            )

# PostgreSQL
# https://www.postgresql.org/docs/current/libpq-envars.html
if 'DATABASES' not in settings and getenv('PGDATABASE'):
    settings['DATABASES'] = {
        'default': {
            'ENGINE': 'django.db.backends.postgresql',
            'NAME': getenv('PGDATABASE'),
            'USER': getenv('PGUSER', ''),
            'PASSWORD': getenv('PGPASSWORD', ''),
            'HOST': getenv('PGHOST', ''),
            'PORT': getenv('PGPORT', ''),
            'TEST': {
                'NAME': 'test_' + getenv('PGDATABASE', ''),
            }
        }
    }

# Make all of the settings attributes of this module
for name, val in settings.items():
    setattr(sys.modules[__name__], name, val)



from typing import Any, Union
import json
from json.decoder import JSONDecodeError

NoneType: type = type(None)


class EnvParser:
    # case insensitive
    bool_true_vals = ('1', 'on', 'true', 'yes')
    bool_false_vals = ('0', 'off', 'false', 'no')
    none_vals = ('none', 'null')

    @classmethod
    def parse(cls, val: str, *expected_types: type) -> Any:
        # Evaluate NoneType first if it's an expected type
        types_ = list(expected_types)
        if NoneType in types_ and types_[0] != NoneType:
            types_.insert(0, types_.pop(types_.index(NoneType)))
        # Attempt to parse val as each type. If all fails raise last ValueError
        for i, type_ in enumerate(types_, start=1):
            try:
                return getattr(cls, f'_{type_.__name__}')(val)
            except ValueError:
                if i < len(types_):
                    pass
                raise

    @classmethod
    def _bool(cls, val: str) -> bool:
        if val.lower() in cls.bool_true_vals:
            return True
        elif val.lower() in cls.bool_false_vals:
            return False
        raise ValueError

    @classmethod
    def _dict(cls, val: str) -> dict:
        dict_ = cls._json(val)
        if not isinstance(dict_, dict):
            raise ValueError(f"Expected dict but received {type(dict_)}")
        return dict_

    @classmethod
    def _float(cls, val: str) -> float:
        return float(val)

    @classmethod
    def _int(cls, val: str) -> int:
        if val.startswith('0o'):
            return int(val, 8)
        elif val.startswith('0x'):
            return int(val, 16)
        return int(val)

    @classmethod
    def _json(cls, val: str) -> Union[dict, list]:
        try:
            return json.loads(val)
        except JSONDecodeError:
            raise ValueError("Invalid JSON")

    @classmethod
    def _list(cls, val: str) -> list:
        if val.startswith('['):
            return cls._json(val)  # type: ignore  # noqa
        return val.split(',')

    @classmethod
    def _NoneType(cls, val: str) -> None:
        if val.lower() in cls.none_vals:
            return
        raise ValueError  # don't need a message

    @classmethod
    def _str(cls, val: str) -> str:
        return val

    @classmethod
    def _tuple(cls, val: str) -> tuple:
        return tuple(cls._list(val))
